using System;
using UnityEngine;
using UnityEngine.UI;
// using System.Collections;
using System.Collections.Generic;
// using SimpleJSON;
using System.Linq;

//  Auto Bulger Plugin by Saking55
//
//  A modified version of the Belly Bulger Plugin by Captain Varghoss.
//
//  It basically works like the Captain Varghoss plugin, but it 
//  selects the target automatically.
//
//  Changes have also been made to the way distances are calculated.
//  Distances are calculated from the Rigid Body that is colliding 
//  rather than the position of the main object or controller.
//
//  Morphings are only applied if the corresponding triggers are in true state.
//
//  Added a smoothing filter to reduce flickering of morphs.
//
//  Morphs converted to Pose Morphs
//  
//  All credit goes to the original creators of these plugins: 
//  VRAdultFun and further enhanced by Captain Varghoss. 
//

namespace Saking55
{
    public class AutoBellyBulger : MVRScript
    {
        protected JSONStorableBool isActive;
        protected JSONStorableBool clampMorph;
        protected JSONStorableBool debugMode;
        protected JSONStorableBool extremeMode;
        protected JSONStorableString uiFocusTarget;
        protected JSONStorableFloat uiTargetOffsetX;
        protected JSONStorableFloat uiTargetOffsetY;
        protected JSONStorableFloat uiTargetOffsetZ;
        protected JSONStorableFloat uiBulgeMult;
        protected JSONStorableFloat uiMinBulgeDistMult;
        protected JSONStorableFloat uiBulge1Mult;
        protected JSONStorableFloat uiBulge2Mult;
        protected JSONStorableFloat uiBulge3Mult;
        protected JSONStorableFloat uiBulge4Mult;
        protected JSONStorableFloat uiBulge5Mult;
        protected JSONStorableFloat uiBulge6Mult;
        protected JSONStorableFloat uiBulge7Mult;
        protected JSONStorableFloat uiBulge8Mult;

        protected JSONStorableFloat uiCurrentDepth;
        protected JSONStorableString helpText;
        protected JSONStorableString helpText2;
        protected JSONStorableString helpText3;

        protected JSONStorableString helpText_Offset;

        protected JSONStorableFloat smoothingFactor;

        protected JSONStorableString filterName;
        protected JSONStorableBool filterNameActive;

        private static float tempFloat = 1.0f;

        private static FreeControllerV3 pelvisController;
        private static DAZMorph morphDeepBulgeBelly1;
        private static DAZMorph morphDeepBulgeBelly2;
        private static DAZMorph morphDeepBulgeBelly3;
        private static DAZMorph morphDeepBulgeBelly4;
        private static DAZMorph morphDeepBulgeBelly5;
        private static DAZMorph morphDeepBulgeBelly6;
        private static DAZMorph morphDeepBulgeBelly7;
        private static DAZMorph morphDeepBulgeBelly8;

        private static Atom person;

        CollisionTriggerEventHandler triggerEvtHdl1;
        CollisionTriggerEventHandler triggerEvtHdl2;
        CollisionTriggerEventHandler triggerEvtHdl3;

        String targetAtomName;
        String selectedTarget;

        private CollisionTrigger triggerVag;
        private CollisionTrigger triggerVagD;
        private CollisionTrigger triggerVagDD;

        private Collider temp;
        Collider colDistTarget = null;
        Dictionary<Collider, bool> colResult;
        long lastColUpdate = 0;
        readonly long timeout = 2000;

        float yVelocity = 0.0f;
        float distance, oldDist, tempDist, minDist = 100f;

        bool notReset = true;
        bool morphsLoaded = true;
        float max = 1f;

        public override void Init()
        {
            try
            {
                colResult = new Dictionary<Collider, bool>();

                // Active Toggle
                isActive = new JSONStorableBool("isActive", true);
                RegisterBool(isActive);
                var toggle = CreateToggle(isActive);
                toggle.label = "Active";

                helpText = new JSONStorableString("Helptext", "\n<b>Auto Belly Bulger</b>\n\nAutomatically selects the collinding object and apply morphs to Person.\n");
                UIDynamic helpTextfield = CreateTextField(helpText, false);
                helpTextfield.height = 210f;

                clampMorph = new JSONStorableBool("clampMorph", true);
                RegisterBool(clampMorph);
                var toggle2 = CreateToggle(clampMorph);
                toggle2.label = "Limit Morphs";

                debugMode = new JSONStorableBool("debugDepth", false);
                RegisterBool(debugMode);
                var toggle3 = CreateToggle(debugMode);
                toggle3.label = "Manual Depth";

                uiFocusTarget = new JSONStorableString("TargetObject", "\n<b>Target:</b> None");
                UIDynamic targetText = CreateTextField(uiFocusTarget, true);
                targetText.height = 100f;

                uiBulgeMult = CreateSlider("Belly Bulge Multiplier", 1f, 0.0f, 3.0f, false, true, false);
                uiMinBulgeDistMult = CreateSlider("Min Bulge Dist Multiplier", 5.0f, 0.0f, 5.0f, false, true, false);
                uiCurrentDepth = CreateSlider("Current Depth", 0.0f, 0f, 1.0f, true, true, false);
                smoothingFactor = CreateSlider("Morph Smoothing", 0.02f, 0f, 1.0f, rightSide: false);

                var btn = CreateButton("Reset Morphs", false);
                btn.button.onClick.AddListener(() => { ResetMorphs(); });

                // Target Atom name filtering
                helpText3 = new JSONStorableString("Helptext3", "\n<b>Target Atom Name Filtering</b>\n\nUse the input below to filter which Atom should be used by the plugin. i.e. Person, Dildo...\nOptional, use if the plugin doesn't select the desired target automatically.");
                UIDynamic helpTextfield3 = CreateTextField(helpText3, false);
                helpTextfield3.height = 280f;

                filterNameActive = new JSONStorableBool("filterNameActive", false);
                RegisterBool(filterNameActive);
                var toggleFilter = CreateToggle(filterNameActive);
                toggleFilter.label = "Filter Atom Name";

                JSONStorableAction ToggleNameFilterActive = new JSONStorableAction("Toggle Name Filter", () =>
                {
                    filterNameActive.val = !filterNameActive.val;
                });
                RegisterAction(ToggleNameFilterActive);


                filterName = new JSONStorableString("filterString", "Person");
                UIDynamic nameFilterText = CreateTextInput(filterName);
                RegisterString(filterName);
                nameFilterText.height = 10f;

                helpText_Offset = new JSONStorableString("helpText_Offset", "\n<b>Target Offset</b>\nOffsets the target position");
                UIDynamic helpTextfield_Offset = CreateTextField(helpText_Offset, false);
                helpTextfield_Offset.height = 120f;
                uiTargetOffsetX = CreateSlider("Target Offset X", 0f, -1.0f, 1.0f, false, true, false);
                uiTargetOffsetY = CreateSlider("Target Offset Y", 0f, -1.0f, 1.0f, false, true, false);
                uiTargetOffsetZ = CreateSlider("Target Offset Z", 0f, -1.0f, 1.0f, false, true, false);

                // Per morph multiplier reducers
                helpText2 = new JSONStorableString("Helptext2", "\n<b>Bulge Morph Adjustments</b>\n\nUse these morphs to control Bulge Multiplier on individual morphs.");
                UIDynamic helpTextfield2 = CreateTextField(helpText2, true);
                helpTextfield2.height = 200f;

                uiBulge1Mult = CreateSlider("Bulge 1 Amount Mult", 0.32f, 0.0f, 1.0f, rightSide: true, constrain: false);
                uiBulge2Mult = CreateSlider("Bulge 2 Amount Mult", 0.67f, 0.0f, 1.0f, rightSide: true, constrain: false);
                uiBulge3Mult = CreateSlider("Bulge 3 Amount Mult", 0.9f, 0.0f, 1.0f, rightSide: true, constrain: false);
                uiBulge4Mult = CreateSlider("Bulge 4 Amount Mult", 1.0f, 0.0f, 1.0f, rightSide: true, constrain: false);
                uiBulge5Mult = CreateSlider("Bulge 5 Amount Mult", 0.86f, 0.0f, 1.0f, rightSide: true, constrain: false);
                uiBulge6Mult = CreateSlider("Bulge 6 Amount Mult", 0.7f, 0.0f, 1.0f, rightSide: true, constrain: false);
                uiBulge7Mult = CreateSlider("Bulge 7 Amount Mult", 0.39f, 0.0f, 1.0f, rightSide: true, constrain: false);
                uiBulge8Mult = CreateSlider("Bulge 8 Amount Mult", 0.18f, 0.0f, 1.0f, rightSide: true, constrain: false);

                extremeMode = new JSONStorableBool("extremeMode", false);
                RegisterBool(extremeMode);
                var btn2 = CreateToggle(extremeMode, true);
                btn2.label = "Extreme Mode";
                // add listener to toggle
                btn2.toggle.onValueChanged.AddListener((bool v) => { ResetMorphs(); });

                // Collisions debug
                // helpText4 = new JSONStorableString("Helptext4", "\n<b>Bulge Morph Adjustments</b>\n\nUse these morphs to reduce Bulge Multiplier on individual morphs.");
                // UIDynamic helpTextfield4 = CreateTextField(helpText4, false);
                // helpTextfield4.height = 200f;

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        private JSONStorableFloat CreateSlider(string sliderName, float defaultValue, float minValue, float maxValue,
            bool constrain = true, bool interactable = true, bool rightSide = false)
        {
            JSONStorableFloat ret = new JSONStorableFloat(sliderName, defaultValue, minValue, maxValue, constrain, interactable);
            RegisterFloat(ret);
            CreateSlider(ret, rightSide);
            return ret;
        }


        void Start()
        {
            try
            {
                person = containingAtom;


                if (person != null)
                {
                    triggerVag = containingAtom.GetStorableByID("VaginaTrigger") as CollisionTrigger;
                    triggerVagD = containingAtom.GetStorableByID("DeepVaginaTrigger") as CollisionTrigger;
                    triggerVagDD = containingAtom.GetStorableByID("DeeperVaginaTrigger") as CollisionTrigger;

                    triggerVag.triggerEnabledJSON.val = true;
                    triggerVagD.triggerEnabledJSON.val = true;
                    triggerVagDD.triggerEnabledJSON.val = true;

                    triggerEvtHdl1 = triggerVag.GetComponent(typeof(CollisionTriggerEventHandler)) as CollisionTriggerEventHandler;
                    triggerEvtHdl2 = triggerVagD.GetComponent(typeof(CollisionTriggerEventHandler)) as CollisionTriggerEventHandler;
                    triggerEvtHdl3 = triggerVagDD.GetComponent(typeof(CollisionTriggerEventHandler)) as CollisionTriggerEventHandler;

                    pelvisController = person.GetStorableByID("chestControl") as FreeControllerV3;
                    ResetMorphs();
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        private void ResetMorphs()
        {
            if (notReset)
            {
                JSONStorable js = person.GetStorableByID("geometry");
                DAZCharacterSelector dcs = js as DAZCharacterSelector;
                GenerateDAZMorphsControlUI morphUI = dcs.morphsControlUI;

                if (morphUI != null)
                {
                    try {
                        morphDeepBulgeBelly1.Reset();
                        morphDeepBulgeBelly2.Reset();
                        morphDeepBulgeBelly3.Reset();
                        morphDeepBulgeBelly4.Reset();
                        morphDeepBulgeBelly5.Reset();
                        morphDeepBulgeBelly6.Reset();
                        morphDeepBulgeBelly7.Reset();
                        morphDeepBulgeBelly8.Reset();  
                    } catch (Exception) {}

                    try
                    {
                        if (extremeMode.val){
                            // SuperController.LogMessage("Extreme Mode on");
                            morphDeepBulgeBelly1 = morphUI.GetMorphByDisplayName("AutoBulger Belly X 1");
                            morphDeepBulgeBelly2 = morphUI.GetMorphByDisplayName("AutoBulger Belly X 2");
                            morphDeepBulgeBelly3 = morphUI.GetMorphByDisplayName("AutoBulger Belly X 3");
                            morphDeepBulgeBelly4 = morphUI.GetMorphByDisplayName("AutoBulger Belly X 4");
                            morphDeepBulgeBelly5 = morphUI.GetMorphByDisplayName("AutoBulger Belly X 5");
                            morphDeepBulgeBelly6 = morphUI.GetMorphByDisplayName("AutoBulger Belly X 6");
                            morphDeepBulgeBelly7 = morphUI.GetMorphByDisplayName("AutoBulger Belly X 7");
                            morphDeepBulgeBelly8 = morphUI.GetMorphByDisplayName("AutoBulger Belly X 8");

                        } else {
                            morphDeepBulgeBelly1 = morphUI.GetMorphByDisplayName("AutoBulger Belly 1");
                            morphDeepBulgeBelly2 = morphUI.GetMorphByDisplayName("AutoBulger Belly 2");
                            morphDeepBulgeBelly3 = morphUI.GetMorphByDisplayName("AutoBulger Belly 3");
                            morphDeepBulgeBelly4 = morphUI.GetMorphByDisplayName("AutoBulger Belly 4");
                            morphDeepBulgeBelly5 = morphUI.GetMorphByDisplayName("AutoBulger Belly 5");
                            morphDeepBulgeBelly6 = morphUI.GetMorphByDisplayName("AutoBulger Belly 6");
                            morphDeepBulgeBelly7 = morphUI.GetMorphByDisplayName("AutoBulger Belly 7");
                            morphDeepBulgeBelly8 = morphUI.GetMorphByDisplayName("AutoBulger Belly 8");
                        }
                            morphDeepBulgeBelly1.Reset();
                            morphDeepBulgeBelly2.Reset();
                            morphDeepBulgeBelly3.Reset();
                            morphDeepBulgeBelly4.Reset();
                            morphDeepBulgeBelly5.Reset();
                            morphDeepBulgeBelly6.Reset();
                            morphDeepBulgeBelly7.Reset();
                            morphDeepBulgeBelly8.Reset();  
                    }
                    catch (Exception)
                    {
                        SuperController.LogError("Morphs not found! \nDisabling AutoBellyBulger.");
                        morphsLoaded = false;
                    }

                }
                notReset = false;
            }
        }

        void FixedUpdate()
        {
            if (isActive.val && morphsLoaded)
            {
                try
                {
                    {  // Find "target"atom and calculate distance
                        targetAtomName = "None";

                        // Find target atom from "vag trigger event" atoms.
                        UpdateCollisionsDict();

                        foreach (KeyValuePair<Collider, bool> pair in colResult.Reverse())
                        {
                            selectedTarget = (pair.Key.attachedRigidbody.GetComponentInParent(typeof(Atom)) as Atom).name;
                            if (selectedTarget.Equals(containingAtom.name))
                            {
                                selectedTarget = "None";
                            }
                            else
                            {
                                if (filterNameActive.val && (selectedTarget.ToLower().Contains(filterName.val.ToLower())) || !filterNameActive.val)
                                {
                                    distance = Vector3.Distance(pelvisController.followWhenOff.position, pair.Key.attachedRigidbody.position);

                                    if (distance < minDist)
                                    {
                                        temp = pair.Key;
                                        minDist = distance;
                                        targetAtomName = selectedTarget;
                                    }
                                }
                            }


                        }

                        colDistTarget = temp;
                        if (targetAtomName == "None")
                        {
                            minDist = 100f;
                        }
                        else
                        {
                            var offset = new Vector3(uiTargetOffsetX.val, uiTargetOffsetY.val, uiTargetOffsetZ.val);
                            // By calculating the rotation Quaternion, align the offset to the target's local coordinate system.
                            oldDist = Vector3.Distance(pelvisController.followWhenOff.position,
                                colDistTarget.attachedRigidbody.position + colDistTarget.attachedRigidbody.rotation * offset);
                            tempDist = Vector3.Distance(pelvisController.followWhenOff.position,
                                temp.attachedRigidbody.position + temp.attachedRigidbody.rotation * offset);

                            minDist = Mathf.Min(tempDist, oldDist);
                        }
                    }  // End Find Target

                    uiFocusTarget.val = "\n<b>Target:</b> " + targetAtomName;     //+ "\ndist: " + tempFloat + " "  + minDist;

                    notReset = true;

                    // Calculate distance between vag and target.
                    if (targetAtomName.Equals("None"))
                        tempFloat = Mathf.SmoothDamp(tempFloat, 1f, ref yVelocity, smoothingFactor.val);
                    else
                    {
                        tempFloat = Mathf.SmoothDamp(tempFloat, Remapp(minDist, 0.23f, 0.43f, 0f, 0.7f), ref yVelocity, smoothingFactor.val);
                    }
                    if (debugMode.val) tempFloat = uiCurrentDepth.val;

                    // Update Morphs
                    CalcUpdateMorph(morphDeepBulgeBelly1, tempFloat, 0.15f, uiBulge1Mult);
                    CalcUpdateMorph(morphDeepBulgeBelly2, tempFloat, 0.13f, uiBulge2Mult);
                    CalcUpdateMorph(morphDeepBulgeBelly3, tempFloat, 0.11f, uiBulge3Mult);
                    CalcUpdateMorph(morphDeepBulgeBelly4, tempFloat, 0.09f, uiBulge4Mult);
                    CalcUpdateMorph(morphDeepBulgeBelly5, tempFloat, 0.07f, uiBulge5Mult);
                    CalcUpdateMorph(morphDeepBulgeBelly6, tempFloat, 0.05f, uiBulge6Mult);
                    CalcUpdateMorph(morphDeepBulgeBelly7, tempFloat, 0.03f, uiBulge7Mult);
                    CalcUpdateMorph(morphDeepBulgeBelly8, tempFloat, 0.025f, uiBulge8Mult);
                    // }
                    if (!debugMode.val) uiCurrentDepth.val = tempFloat;
                    minDist = 10f;
                }
                catch (Exception e)
                {
                    SuperController.LogError("Exception: " + e);
                }
            }
            else { ResetMorphs(); }
        }

        private void CalcUpdateMorph(DAZMorph targetMorph, float distance, float eachDist0, JSONStorableFloat eachBulgeMult)
        {
            targetMorph.morphValue =
                Mathf.Max((1.0f - distance / (eachDist0 * uiMinBulgeDistMult.val)) * uiBulgeMult.val * eachBulgeMult.val, 0.0f);
            if (clampMorph.val) targetMorph.morphValue = Mathf.Clamp(targetMorph.morphValue, 0f, max);
        }

        private float Remapp(float val, float from1, float to1, float from2, float to2)
        {
            return (val - from1) / (to1 - from1) * (to2 - from2) + from2;
        }

        void OnDestroy()
        {
            ResetMorphs();
        }

        void OnDisable()
        {
            ResetMorphs();
        }

        private void UpdateCollisionsDict()
        {
            var now = DateTime.Now.Ticks / TimeSpan.TicksPerMillisecond;

            if (triggerVag.trigger.active || triggerVagD.trigger.active || triggerVagDD.trigger.active)
            {
                // while collisions are happening, keeps track of all past collisions
                colResult = triggerEvtHdl1.collidingWithDictionary
                            .Concat(triggerEvtHdl2.collidingWithDictionary)
                            .Concat(triggerEvtHdl3.collidingWithDictionary).Concat(colResult)
                            .GroupBy(d => d.Key).ToDictionary(d => d.Key, d => d.First().Value);

                lastColUpdate = now;

            }
            else if (now > lastColUpdate + timeout)
            {
                // dumps collisions after x ms of no collision
                colResult.Clear();
            }

            var itemsToRemove = colResult.Where(f => f.Key.attachedRigidbody == null).ToArray();
            foreach (var item in itemsToRemove)
            {
                colResult.Remove(item.Key);
                // SuperController.LogError("Removed");
            }


        }

        public UIDynamicTextField CreateTextInput(JSONStorableString jss, bool rightSide = false)
        {
            var textfield = CreateTextField(jss, rightSide);
            textfield.height = 20f;
            textfield.backgroundColor = Color.white;
            var input = textfield.gameObject.AddComponent<InputField>();
            var rect = input.GetComponent<RectTransform>().sizeDelta = new Vector2(1f, 0.4f);
            input.textComponent = textfield.UItext;
            jss.inputField = input;
            return textfield;
        }

    }
}

