using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;

//  Auto Bulger Plugin by Saking55
//
//  A modified version of the Throat Bulger Plugin by Captain Varghoss.
//
//  It basically works like the Captain Varghoss plugin, but it 
//  selects the target automatically.
//
//  Changes have also been made to the way distances are calculated.
//  Distances are calculated from the Rigid Body that is colliding 
//  rather than the position of the main object or controller.
//
//  Morphings are only applied if the corresponding triggers are in true state.
//
//  Added a smoothing filter to reduce flickering of morphs.
//
//  Morphs converted to Pose Morphs
//
//  All credit goes to the original creators of these plugins: 
//  VRAdultFun and further enhanced by Captain Varghoss. 
//

namespace Saking55
{
    public class AutoThroatBulger : MVRScript
    {
        protected JSONStorableBool isActive;
        protected JSONStorableBool clampMorph;
        protected JSONStorableBool debugMode;
        protected JSONStorableString uiFocusTarget;
        protected JSONStorableFloat uiBulgeMult;
        protected JSONStorableFloat uiMinBulgeDistMult;
        protected JSONStorableFloat uiBulge1Mult;
        protected JSONStorableFloat uiBulge2Mult;
        protected JSONStorableFloat uiBulge3Mult;
        protected JSONStorableFloat uiBulge4Mult;

        protected JSONStorableFloat uiTargetOffsetX;
        protected JSONStorableFloat uiTargetOffsetY;
        protected JSONStorableFloat uiTargetOffsetZ;

        protected JSONStorableFloat uiCurrentDepth;
        protected JSONStorableString helpText;
        protected JSONStorableString helpText2;
        protected JSONStorableString helpText3;

        protected JSONStorableString helpText_Offset;

        protected JSONStorableFloat smoothingFactor;

        protected JSONStorableString filterName;
        protected JSONStorableBool filterNameActive;

        private static float tempFloat = 1.0f;

        private static FreeControllerV3 neckController;
        private static DAZMorph morphDeepBulgeThroat1;
        private static DAZMorph morphDeepBulgeThroat2;
        private static DAZMorph morphDeepBulgeThroat3;
        private static DAZMorph morphDeepBulgeThroat4;

        private static Atom person;
        // private static Atom currentAtom;

        private CollisionTriggerEventHandler triggerEvtHdl1;
        private CollisionTriggerEventHandler triggerEvtHdl2;

        String targetAtomName;
        String selectedTarget;

        private CollisionTrigger triggerMouth;
        private CollisionTrigger triggerThroat;

        private Collider temp;
        private Collider colDistTarget = null;
        Dictionary<Collider, bool> colResult;
        long lastColUpdate = 0;

        float yVelocity = 0.0f;
        float distance, oldDist, tempDist, minDist = 10f;

        bool notReset = true;
        bool morphsLoaded = true;
        readonly float max = 1f; // max morphing value



        public override void Init()
        {
            try
            {
                // Active Toggle
                isActive = new JSONStorableBool("isActive", true);
                RegisterBool(isActive);
                var toggle = CreateToggle(isActive);
                toggle.label = "Active";

                colResult = new Dictionary<UnityEngine.Collider, bool>();

                JSONStorableAction ToggleActive = new JSONStorableAction("Toggle Throat Bulger", () =>
                {
                    isActive.val = !isActive.val;
                });
                RegisterAction(ToggleActive);

                helpText = new JSONStorableString("Helptext", "\n<b>Auto Throat Bulger</b>\n\nAutomatically selects the collinding object and apply morphs to Person.\n");
                UIDynamic helpTextfield = CreateTextField(helpText, false);
                helpTextfield.height = 210f;

                uiFocusTarget = new JSONStorableString("TargetObject", "\n<b>Target:</b> None");
                UIDynamic targetText = CreateTextField(uiFocusTarget, true);
                targetText.height = 100f;
                //helpTextfield.height = 700f;

                clampMorph = new JSONStorableBool("clampMorph", true);
                RegisterBool(clampMorph);
                var toggle2 = CreateToggle(clampMorph);
                toggle2.label = "Limit Morphs";

                debugMode = new JSONStorableBool("debugDepth", false);
                RegisterBool(debugMode);
                var toggle3 = CreateToggle(debugMode);
                toggle3.label = "Manual Depth";



                uiBulgeMult = CreateSlider("Throat Bulge Multiplier", 1f, 0.0f, 1.0f, rightSide: false, constrain: false);
                uiMinBulgeDistMult = CreateSlider("Min Bulge Dist Multiplier", 2f, 0.0f, 5.0f, rightSide: false, constrain: false);
                uiCurrentDepth = CreateSlider("Current Depth", 1f, 0.0f, 1.0f, rightSide: false, constrain: false);
                smoothingFactor = CreateSlider("Morph Smoothing", 0.1f, 0.0f, 1.0f, rightSide: false, constrain: false);

                var btn = CreateButton("Reset Morphs", false);
                btn.button.onClick.AddListener(() => { ResetMorphs(); });




                // Per morph multiplier reducers
                helpText2 = new JSONStorableString("Helptext2", "\n<b>Bulge Morph Adjustments</b>\n\nUse these morphs to control Bulge Multiplier on individual morphs.");
                UIDynamic helpTextfield2 = CreateTextField(helpText2, true);
                helpTextfield2.height = 200f;


                uiBulge1Mult = CreateSlider("Bulge 1 Amount Mult", 1.0f, 0.0f, 1.0f, rightSide: true, constrain: false);
                uiBulge2Mult = CreateSlider("Bulge 2 Amount Mult", 1.0f, 0.0f, 1.0f, rightSide: true, constrain: false);
                uiBulge3Mult = CreateSlider("Bulge 3 Amount Mult", 1.0f, 0.0f, 1.0f, rightSide: true, constrain: false);
                uiBulge4Mult = CreateSlider("Bulge 4 Amount Mult", 0.5f, 0.0f, 1.0f, rightSide: true, constrain: false);

                // Target Atom name filtering
                helpText3 = new JSONStorableString("Helptext3", "\n<b>Target Atom Name Filtering</b>\n\nUse the input below to filter which Atom should be used by the plugin. i.e. Person, Dildo...\nOptional, use if the plugin doesn't select the desired target automatically.");
                UIDynamic helpTextfield3 = CreateTextField(helpText3, true);
                helpTextfield3.height = 280f;

                filterNameActive = new JSONStorableBool("filterNameActive", false);
                RegisterBool(filterNameActive);
                var toggleFilter = CreateToggle(filterNameActive, true);
                toggleFilter.label = "Filter Atom Name";

                JSONStorableAction ToggleNameFilterActive = new JSONStorableAction("Toggle Name Filter", () =>
                {
                    filterNameActive.val = !filterNameActive.val;
                });
                RegisterAction(ToggleNameFilterActive);


                filterName = new JSONStorableString("filterString", "Person");
                UIDynamic nameFilterText = CreateTextInput(filterName, true);
                RegisterString(filterName);

                nameFilterText.height = 10f;

                helpText_Offset = new JSONStorableString("helpText_Offset", "\n<b>Target Offset</b>\nOffsets the target position");
                UIDynamic helpTextfield_Offset = CreateTextField(helpText_Offset, false);
                helpTextfield_Offset.height = 120f;
                uiTargetOffsetX = CreateSlider("Target Offset X", 0f, -1.0f, 1.0f, true, true, false);
                uiTargetOffsetY = CreateSlider("Target Offset Y", 0f, -1.0f, 1.0f, true, true, false);
                uiTargetOffsetZ = CreateSlider("Target Offset Z", 0f, -1.0f, 1.0f, true, true, false);


                // Collisions debug
                // helpText4 = new JSONStorableString("Helptext4", "\n<b>Collisions</b>\n\nNone");
                // UIDynamic helpTextfield4 = CreateTextField(helpText4, false);
                // helpTextfield4.height = 200f;
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        void Start()
        {
            try
            {
                person = containingAtom;
                if (person != null)
                {
                    triggerMouth = (containingAtom.GetStorableByID("MouthTrigger") as CollisionTrigger);
                    triggerThroat = (containingAtom.GetStorableByID("ThroatTrigger") as CollisionTrigger);

                    triggerMouth.triggerEnabledJSON.val = true;
                    triggerThroat.triggerEnabledJSON.val = true;

                    triggerEvtHdl1 = triggerMouth.GetComponent(typeof(CollisionTriggerEventHandler)) as CollisionTriggerEventHandler;
                    triggerEvtHdl2 = triggerThroat.GetComponent(typeof(CollisionTriggerEventHandler)) as CollisionTriggerEventHandler;

                    neckController = person.GetStorableByID("neckControl") as FreeControllerV3;
                    ResetMorphs();
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        private void ResetMorphs()
        {
            if (notReset)
            {
                JSONStorable js = person.GetStorableByID("geometry");
                DAZCharacterSelector dcs = js as DAZCharacterSelector;
                GenerateDAZMorphsControlUI morphUI = dcs.morphsControlUI;

                if (morphUI != null)
                {
                    try
                    {
                        morphDeepBulgeThroat1 = morphUI.GetMorphByDisplayName("AutoBulger Throat 1");
                        morphDeepBulgeThroat1.Reset();
                        morphDeepBulgeThroat2 = morphUI.GetMorphByDisplayName("AutoBulger Throat 2");
                        morphDeepBulgeThroat2.Reset();
                        morphDeepBulgeThroat3 = morphUI.GetMorphByDisplayName("AutoBulger Throat 3");
                        morphDeepBulgeThroat3.Reset();
                        morphDeepBulgeThroat4 = morphUI.GetMorphByDisplayName("AutoBulger Throat 4");
                        morphDeepBulgeThroat4.Reset();
                    }
                    catch (Exception)
                    {
                        SuperController.LogError("Morphs not found! \nDisabling AutoThroatBulger.");
                        morphsLoaded = false;
                    }
                }
                notReset = false;
            }
        }

        void FixedUpdate()
        {
            if (isActive.val && morphsLoaded)
            {
                try
                {
                    // helpText4.val = "<b>Collisions:</b>";
                    targetAtomName = "None";

                    UpdateCollisionsDict();

                    foreach (KeyValuePair<Collider, bool> pair in colResult)
                    {
                        selectedTarget = (pair.Key.attachedRigidbody.GetComponentInParent(typeof(Atom)) as Atom).name;
                        if (selectedTarget.Equals(containingAtom.name))
                        {
                            selectedTarget = "None";
                        }
                        else
                        {
                            if (filterNameActive.val && (selectedTarget.ToLower().Contains(filterName.val.ToLower())) || !filterNameActive.val)
                            {
                                // distance = remapp(Vector3.Distance(neckController.followWhenOff.position, pair.Key.attachedRigidbody.position), 0.085f, 0.16f, 0f, 0.3f);
                                // helpText4.val += "\n<b>" + selectedTarget + " - distance:</b> "+ distance;
                                distance = Vector3.Distance(neckController.followWhenOff.position, pair.Key.attachedRigidbody.position);

                                if (distance < minDist)
                                {
                                    temp = pair.Key;
                                    minDist = distance;
                                    targetAtomName = selectedTarget;
                                }
                            }
                            // break;   
                        }
                    }

                    if (targetAtomName == "None")
                    {
                        colDistTarget = null;
                        minDist = 10f;
                    }

                    if (colDistTarget == null)
                    {
                        colDistTarget = temp;
                    }
                    else
                    {
                        var offset = new Vector3(uiTargetOffsetX.val, uiTargetOffsetY.val, uiTargetOffsetZ.val);
                        // By calculating the rotation Quaternion, align the offset to the target's local coordinate system.
                        oldDist = Vector3.Distance(neckController.followWhenOff.position,
                            colDistTarget.attachedRigidbody.position + colDistTarget.attachedRigidbody.rotation * offset);
                        tempDist = Vector3.Distance(neckController.followWhenOff.position,
                            temp.attachedRigidbody.position + temp.attachedRigidbody.rotation * offset);

                        // oldDist =  Vector3.Distance(neckController.followWhenOff.position, colDistTarget.attachedRigidbody.position);
                        // tempDist = Vector3.Distance(neckController.followWhenOff.position, temp.attachedRigidbody.position);

                        if (tempDist < oldDist)
                            minDist = tempDist;
                        else
                            minDist = oldDist;

                    }

                    uiFocusTarget.val = "\n<b>Target:</b> " + targetAtomName;//  + "\ndist: " + tempFloat + " "  + minDist;

                    // if (morphDeepBulgeThroat1 != null && morphDeepBulgeThroat2 != null && morphDeepBulgeThroat3 != null)
                    // {
                    notReset = true;
                    if (targetAtomName.Equals("None"))
                        tempFloat = Mathf.SmoothDamp(tempFloat, 1f, ref yVelocity, smoothingFactor.val);
                    else
                    {
                        //helpText.val += "Distance: <b>" + Vector3.Distance(neckController.followWhenOff.position, colDistTarget.attachedRigidbody.position) + "</b>";
                        tempFloat = Mathf.SmoothDamp(tempFloat, Remapp(minDist, 0.07f, 0.19f, 0f, 0.8f), ref yVelocity, smoothingFactor.val);
                    }
                    if (debugMode.val) tempFloat = uiCurrentDepth.val;

                    CalcUpdateMorph(morphDeepBulgeThroat1, tempFloat, 0.4f, uiBulge1Mult);
                    CalcUpdateMorph(morphDeepBulgeThroat2, tempFloat, 0.2f, uiBulge2Mult);
                    CalcUpdateMorph(morphDeepBulgeThroat3, tempFloat, 0.15f, uiBulge3Mult);
                    CalcUpdateMorph(morphDeepBulgeThroat4, tempFloat, 0.4f, uiBulge4Mult);

                    // }
                    if (!debugMode.val) uiCurrentDepth.val = tempFloat;
                    minDist = 10f;
                }
                catch (Exception e)
                {
                    SuperController.LogError("Exception caught: " + e);
                }
            }
            else { ResetMorphs(); }
        }


        private JSONStorableFloat CreateSlider(string sliderName, float defaultValue, float minValue, float maxValue,
            bool constrain = true, bool interactable = true, bool rightSide = false)
        {
            JSONStorableFloat ret = new JSONStorableFloat(sliderName, defaultValue, minValue, maxValue, constrain, interactable);
            RegisterFloat(ret);
            CreateSlider(ret, rightSide);
            return ret;
        }

        private void CalcUpdateMorph(DAZMorph targetMorph, float distance, float eachDist0, JSONStorableFloat eachBulgeMult)
        {
            targetMorph.morphValue = Mathf.Max((1.0f - distance / (eachDist0 * uiMinBulgeDistMult.val))
                * uiBulgeMult.val * eachBulgeMult.val, 0.0f);
            if (clampMorph.val) targetMorph.morphValue = Mathf.Clamp(targetMorph.morphValue, 0f, max);
        }

        private void UpdateCollisionsDict()
        {
            var now = DateTime.Now.Ticks / TimeSpan.TicksPerMillisecond;
            if (triggerMouth.trigger.active || triggerThroat.trigger.active)
            {
                // while collisions are happening, keeps track of all past collisions
                colResult = triggerEvtHdl1.collidingWithDictionary
                    .Concat(triggerEvtHdl2.collidingWithDictionary).Concat(colResult)
                    .GroupBy(d => d.Key).ToDictionary(d => d.Key, d => d.First().Value);

                lastColUpdate = now;

            }
            else if (now > lastColUpdate + 500)
            {
                // dumps collisions after 500 ms of no collision
                colResult.Clear();
            }

            var itemsToRemove = colResult.Where(f => f.Key.attachedRigidbody == null).ToArray();
            foreach (var item in itemsToRemove)
            {
                colResult.Remove(item.Key);
                // SuperController.LogError("Removed");
            }


        }

        private float Remapp(float val, float from1, float to1, float from2, float to2)
        {
            return (val - from1) / (to1 - from1) * (to2 - from2) + from2;
        }

        void OnDestroy()
        {
            ResetMorphs();
        }

        void OnDisable()
        {
            ResetMorphs();
        }

        public UIDynamicTextField CreateTextInput(JSONStorableString jss, bool rightSide = false)
        {
            var textfield = CreateTextField(jss, rightSide);
            textfield.height = 20f;
            textfield.backgroundColor = Color.white;
            var input = textfield.gameObject.AddComponent<InputField>();
            var rect = input.GetComponent<RectTransform>().sizeDelta = new Vector2(1f, 0.4f);
            input.textComponent = textfield.UItext;
            jss.inputField = input;
            return textfield;
        }

    }
}